/*  CodeEdit -:-  (C)opyright 2001-2002 Andreas Engh-Halstvedt
 *  AEdit -:-     (C)opyright 2000-2002 Kristian Van Der Vliet
 *  AEdit -:-     (C)opyright 2004-2006 Jonas Jarvoll
 *  CodeEdit -:-  (C)opyright 2011-2012 David Kent
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details. 
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <util/application.h>
#include <util/resources.h>
#include <gui/font.h>
#include <gui/image.h>
#include <gui/popupmenu.h>
#include <gui/textview.h>
#include <gui/requesters.h>

#include <fstream>

#include <appwindow.h>
#include <messages.h>
#include <version.h>
#include <buffer.h>
#include <editview.h>


AppWindow::AppWindow() : os::Window( os::Rect(100,125,700,500), "app_window", "" )
{
	// Set pointers to a well defined value
	pcCurrentBuffer = NULL;
	pcFindDialog = NULL;
	pcReplaceDialog = NULL;
	pcGotoDialog = NULL;
	pcFontRequester = NULL;
	pcCurrentDialog = NULL;
	pcLoadRequester = NULL;
	pcSaveRequester = NULL;
	pcSaveAsRequester = NULL;

	// Load settings
	cSettings.Load();

	// Set up latest path used
	LatestOpenPath = cSettings.GetString("LoadPath");
	LatestSavePath = cSettings.GetString("SavePath");

	// Make all menus
	SetupMenus();

	// Make the buttons & attach to a toolbar
	SetupButtonBar();

	// Create & attach the editor view
	pcEditorView = new MyTabView( os::Rect(), this );
	pcEditorView->SetTarget(this);
	pcEditorView->SetMessage( new os::Message(M_INVOKED_TAB_CHANGED) );
	AddChild(pcEditorView);

	// Create the status bar
	pcStatusBar = new os::StatusBar( os::Rect(), "appwindow_status_bar" );
	pcStatusBar->AddPanel( "message", "" );
	AddChild(pcStatusBar);

	// Fix layout by setting the position of the window
	SetWindowPosition();

	// Set the window title etc.
	SetSizeLimits( os::Point(400,300), os::Point(4096,4096) );

	// Update menu items
	UpdateMenuItemStatus();

	// Set up view
	pcViewToolbar->SetChecked( cSettings.GetBool("tool_bar", true) );
	pcViewStatusbar->SetChecked( cSettings.GetBool("status_bar", true) );

	// Set an application icon for Dock
	os::Resources cRes( get_image_id() );
	os::BitmapImage* pcImage = new os::BitmapImage();
	pcImage->Load( cRes.GetResourceStream("icon48x48.png") );
	SetIcon( pcImage->LockBitmap() );
}

AppWindow::~AppWindow()
{
	delete(pcFindDialog);
	delete(pcReplaceDialog);
	delete(pcGotoDialog);
}

void AppWindow::FrameSized( const os::Point& cDelta )
{
	os::Window::FrameSized(cDelta);
	Layout();
}

void AppWindow::SetupMenus()
{
	// Create a pointer to the 16x16 menubar images
	os::Resources cRes( get_image_id() );
	os::BitmapImage* pcMenuBarIcon;

	// Create the menu bar
	pcMenuBar = new os::Menu( os::Rect(), "main_menu", os::ITEMS_IN_ROW, os::CF_FOLLOW_LEFT | os::CF_FOLLOW_RIGHT | os::CF_FOLLOW_TOP );

	// Create the application menu
	pcAppMenu = new os::Menu( os::Rect(), "Application", os::ITEMS_IN_COLUMN );
	pcMenuBarIcon = new os::BitmapImage();
	pcMenuBarIcon->Load( cRes.GetResourceStream("about16x16.png") );
	pcAppMenu->AddItem( "About", new os::Message(M_MENU_APP_ABOUT), "", pcMenuBarIcon );
	pcAppMenu->AddItem( new os::MenuSeparator() );
	pcMenuBarIcon = new os::BitmapImage();
	pcMenuBarIcon->Load( cRes.GetResourceStream("exit16x16.png") );
	pcAppMenu->AddItem( "Quit", new os::Message(M_MENU_APP_QUIT), "", pcMenuBarIcon );
	pcMenuBar->AddItem(pcAppMenu);

	// Create the file menu
	pcMenuBarIcon = new os::BitmapImage();
	pcMenuBarIcon->Load( cRes.GetResourceStream("filenew16x16.png") );
	pcFileMenu = new os::Menu( os::Rect(), "File", os::ITEMS_IN_COLUMN );
	pcFileMenu->AddItem( new os::MenuItem( "New", new os::Message(M_MENU_FILE_NEW), "", pcMenuBarIcon ) );
	pcMenuBarIcon = new os::BitmapImage();
	pcMenuBarIcon->Load( cRes.GetResourceStream("fileopen16x16.png") );
	pcFileMenu->AddItem(new os::MenuItem( "Open...", new os::Message(M_MENU_FILE_OPEN), "", pcMenuBarIcon ));
	pcMenuBarIcon = new os::BitmapImage();
	pcMenuBarIcon->Load( cRes.GetResourceStream("fileclose16x16.png") );
	pcFileClose = new os::MenuItem( "Close", new os::Message(M_MENU_FILE_CLOSE), "", pcMenuBarIcon );
	pcFileMenu->AddItem(pcFileClose);
	pcFileMenu->AddItem( new os::MenuSeparator() );
	pcMenuBarIcon = new os::BitmapImage();
	pcMenuBarIcon->Load( cRes.GetResourceStream("filesave16x16.png") );
	pcFileSave = new os::MenuItem( "Save", new os::Message(M_MENU_FILE_SAVE), "", pcMenuBarIcon );
	pcFileMenu->AddItem(pcFileSave);
	pcMenuBarIcon = new os::BitmapImage();
	pcMenuBarIcon->Load( cRes.GetResourceStream("filesaveas16x16.png") );
	pcFileSaveAs = new os::MenuItem( "Save as...", new os::Message(M_MENU_FILE_SAVE_AS), "", pcMenuBarIcon );
	pcFileMenu->AddItem(pcFileSaveAs);
	pcMenuBar->AddItem(pcFileMenu);

	// Create the edit menu
	pcMenuBarIcon = new os::BitmapImage();
	pcMenuBarIcon->Load( cRes.GetResourceStream("editcut16x16.png") );
	pcEditMenu = new os::Menu( os::Rect(), "Edit", os::ITEMS_IN_COLUMN );
	pcEditCut = new os::MenuItem( "Cut", new os::Message(M_MENU_EDIT_CUT), "", pcMenuBarIcon );
	pcEditMenu->AddItem(pcEditCut);
	pcMenuBarIcon = new os::BitmapImage();
	pcMenuBarIcon->Load( cRes.GetResourceStream("editcopy16x16.png") );
	pcEditCopy = new os::MenuItem( "Copy",new os::Message(M_MENU_EDIT_COPY), "", pcMenuBarIcon );
	pcEditMenu->AddItem(pcEditCopy);
	pcMenuBarIcon = new os::BitmapImage();
	pcMenuBarIcon->Load( cRes.GetResourceStream("editpaste16x16.png") );
	pcEditPaste = new os::MenuItem( "Paste", new os::Message(M_MENU_EDIT_PASTE), "", pcMenuBarIcon );
	pcEditMenu->AddItem(pcEditPaste);
	pcEditMenu->AddItem(new os::MenuSeparator());
	pcMenuBarIcon = new os::BitmapImage();
	pcMenuBarIcon->Load( cRes.GetResourceStream("editselectall16x16.png") );
	pcEditSelectAll = new os::MenuItem( "Select All", new os::Message(M_MENU_EDIT_SELECT_ALL), "", pcMenuBarIcon );
	pcEditMenu->AddItem(pcEditSelectAll);
	pcEditMenu->AddItem( new os::MenuSeparator() );
	pcMenuBarIcon = new os::BitmapImage();
	pcMenuBarIcon->Load( cRes.GetResourceStream("editundo16x16.png") );
	pcEditUndo = new os::MenuItem( "Undo", new os::Message(M_MENU_EDIT_UNDO), "", pcMenuBarIcon );
	pcEditMenu->AddItem(pcEditUndo);
	pcMenuBarIcon = new os::BitmapImage();
	pcMenuBarIcon->Load( cRes.GetResourceStream("editredo16x16.png") );
	pcEditRedo = new os::MenuItem( "Redo", new os::Message(M_MENU_EDIT_REDO), "", pcMenuBarIcon );
	pcEditMenu->AddItem(pcEditRedo);
	pcMenuBar->AddItem(pcEditMenu);

	// Create the view menu
	pcViewMenu = new os::Menu( os::Rect(), "View", os::ITEMS_IN_COLUMN );
	pcViewToolbar = new os::CheckMenu( "tool_bar", new os::Message(M_MENU_VIEW_TOOLBAR), "" );
	pcViewToolbar->SetChecked(true);
	pcViewMenu->AddItem(pcViewToolbar);
	pcViewStatusbar = new os::CheckMenu( "status_bar", new os::Message(M_MENU_VIEW_STATUSBAR), "" );
	pcViewStatusbar->SetChecked(true);
	pcViewMenu->AddItem(pcViewStatusbar);
	pcViewMenu->AddItem( new os::MenuSeparator() );
	pcMenuBarIcon = new os::BitmapImage();
	pcMenuBarIcon->Load( cRes.GetResourceStream("viewfont16x16.png") );
	pcViewMenu->AddItem( new os::MenuItem( "Set Font...", new os::Message(M_MENU_VIEW_SETFONT), "", pcMenuBarIcon ) );
	pcMenuBar->AddItem(pcViewMenu);

	// Create the find menu
	pcFindMenu = new os::Menu( os::Rect(), "Find", os::ITEMS_IN_COLUMN );
	pcMenuBarIcon = new os::BitmapImage();
	pcMenuBarIcon->Load( cRes.GetResourceStream("findsearch16x16.png") );
	pcFindMenu->AddItem( "Find...", new os::Message(M_MENU_FIND_FIND), "", pcMenuBarIcon );
	pcMenuBarIcon = new os::BitmapImage();
	pcMenuBarIcon->Load( cRes.GetResourceStream("findreplace16x16.png") );
	pcFindMenu->AddItem( "Replace...", new os::Message(M_MENU_FIND_REPLACE), "", pcMenuBarIcon );
	pcFindMenu->AddItem( new os::MenuSeparator() );
	pcMenuBarIcon = new os::BitmapImage();
	pcMenuBarIcon->Load( cRes.GetResourceStream("findjumpto16x16.png") );
	pcFindMenu->AddItem( "Go to Line...", new os::Message(M_MENU_FIND_GOTO), "", pcMenuBarIcon );
	pcMenuBar->AddItem(pcFindMenu);

	// Create the settings menu
	pcSettingsMenu = new os::Menu( os::Rect(0,0,1,1), "Settings", os::ITEMS_IN_COLUMN );
	pcSettingsSaveOnClose = new os::CheckMenu( "Save Settings on Close", new os::Message(M_MENU_SETTINGS_SAVE_ON_CLOSE), cSettings.GetBool("SaveOnExit", true) );
	pcSettingsMenu->AddItem(pcSettingsSaveOnClose);
	pcSettingsMenu->AddItem( new os::MenuSeparator() );
	pcSettingsMenu->AddItem( "Save Settings Now", new os::Message(M_MENU_SETTINGS_SAVE_NOW) );
	pcSettingsMenu->AddItem( "Reset Settings", new os::Message(M_MENU_SETTINGS_RESET) );
	pcMenuBar->AddItem(pcSettingsMenu);

	// Create the help menu
	pcHelpMenu = new os::Menu( os::Rect(), "Help", os::ITEMS_IN_COLUMN );
	pcMenuBarIcon = new os::BitmapImage();
	pcMenuBarIcon->Load( cRes.GetResourceStream("about16x16.png") );
	pcHelpMenu->AddItem( "About CodeEdit", new os::Message(M_MENU_APP_ABOUT), "", pcMenuBarIcon );
	pcHelpMenu->AddItem( new os::MenuSeparator() );
	pcMenuBarIcon = new os::BitmapImage();
	pcMenuBarIcon->Load( cRes.GetResourceStream("settings16x16.png") );
	pcHelpMenu->AddItem( "Keyboard Shortcuts", new os::Message(M_MENU_HELP_SHORTCUTS), "", pcMenuBarIcon );
	pcMenuBar->AddItem(pcHelpMenu);

	// Attach the menu bar to the window
	pcMenuBar->SetTargetForItems(this);
	AddChild(pcMenuBar);
}

void AppWindow::SetupButtonBar()
{
	// Create a pointer to the 16x16 toolbar images
	os::Resources cRes( get_image_id() );
	os::BitmapImage* pcToolBarIcon;

	// Create the button bar
	pcToolBar = new os::ToolBar( os::Rect(), "button_bar" );
	pcToolBar->AddSeparator("");

	// Create the new button
	pcToolBarIcon = new os::BitmapImage();
	pcToolBarIcon->Load( cRes.GetResourceStream("filenew16x16.png") );
	pcImageFileNew = new os::ImageButton( os::Rect(), "appwindow_filenew", "New", new os::Message(M_BUT_FILE_NEW), pcToolBarIcon, os::ImageButton::IB_TEXT_BOTTOM, true, true, true );
	pcToolBar->AddChild( pcImageFileNew, os::ToolBar::TB_FIXED_WIDTH );

	// Create the open button
	pcToolBarIcon = new os::BitmapImage();
	pcToolBarIcon->Load( cRes.GetResourceStream("fileopen16x16.png") );
	pcImageFileOpen = new os::ImageButton( os::Rect(), "appwindow_fileopen", "Open...", new os::Message(M_BUT_FILE_OPEN), pcToolBarIcon, os::ImageButton::IB_TEXT_BOTTOM, true, true, true );
	pcToolBar->AddChild( pcImageFileOpen, os::ToolBar::TB_FIXED_WIDTH );

	// Create the save button
	pcToolBarIcon = new os::BitmapImage();
	pcToolBarIcon->Load( cRes.GetResourceStream("filesave16x16.png") );
	pcImageFileSave = new os::ImageButton( os::Rect(), "appwindow_filesave", "Save", new os::Message(M_BUT_FILE_SAVE), pcToolBarIcon, os::ImageButton::IB_TEXT_BOTTOM, true, true, true );
	pcToolBar->AddChild( pcImageFileSave, os::ToolBar::TB_FIXED_WIDTH );

	// Create the saveas button
	pcToolBarIcon = new os::BitmapImage();
	pcToolBarIcon->Load( cRes.GetResourceStream("filesaveas16x16.png") );
	pcImageFileSaveAs = new os::ImageButton( os::Rect(), "appwindow_filesaveas", "Save As", new os::Message(M_BUT_FILE_SAVE_AS), pcToolBarIcon, os::ImageButton::IB_TEXT_BOTTOM, true, true, true );
	pcToolBar->AddChild( pcImageFileSaveAs, os::ToolBar::TB_FIXED_WIDTH );

	// Create the close button
	pcToolBarIcon = new os::BitmapImage();
	pcToolBarIcon->Load( cRes.GetResourceStream("fileclose16x16.png") );
	pcImageFileClose = new os::ImageButton( os::Rect(), "appwindow_fileclose", "Close", new os::Message(M_BUT_FILE_CLOSE), pcToolBarIcon, os::ImageButton::IB_TEXT_BOTTOM, true, true, true );
	pcToolBar->AddChild( pcImageFileClose, os::ToolBar::TB_FIXED_WIDTH );

	pcToolBar->AddSeparator("");

	// Create the cut button
	pcToolBarIcon = new os::BitmapImage();
	pcToolBarIcon->Load( cRes.GetResourceStream("editcut16x16.png") );
	pcImageEditCut = new os::ImageButton( os::Rect(), "appwindow_editcut", "Cut", new os::Message(M_BUT_EDIT_CUT), pcToolBarIcon, os::ImageButton::IB_TEXT_BOTTOM, true, true, true );
	pcToolBar->AddChild( pcImageEditCut, os::ToolBar::TB_FIXED_WIDTH );

	// Create the copy button
	pcToolBarIcon = new os::BitmapImage();
	pcToolBarIcon->Load( cRes.GetResourceStream("editcopy16x16.png") );
	pcImageEditCopy = new os::ImageButton( os::Rect(), "appwindow_editcopy", "Copy", new os::Message(M_BUT_EDIT_COPY), pcToolBarIcon, os::ImageButton::IB_TEXT_BOTTOM, true, true, true );	
	pcToolBar->AddChild( pcImageEditCopy, os::ToolBar::TB_FIXED_WIDTH );

	// Create the paste button
	pcToolBarIcon = new os::BitmapImage();
	pcToolBarIcon->Load( cRes.GetResourceStream("editpaste16x16.png") );
	pcImageEditPaste = new os::ImageButton( os::Rect(), "appwindow_editpaste", "Paste", new os::Message(M_BUT_EDIT_PASTE), pcToolBarIcon, os::ImageButton::IB_TEXT_BOTTOM, true, true, true );	
	pcToolBar->AddChild( pcImageEditPaste, os::ToolBar::TB_FIXED_WIDTH );

	pcToolBar->AddSeparator("");

	// Create the selectall button
	pcToolBarIcon = new os::BitmapImage();
	pcToolBarIcon->Load( cRes.GetResourceStream("editselectall16x16.png") );
	pcImageEditSelectAll = new os::ImageButton( os::Rect(), "appwindow_selectall", "Select All", new os::Message(M_BUT_EDIT_SELECT_ALL), pcToolBarIcon, os::ImageButton::IB_TEXT_BOTTOM, true, true, true );
	pcToolBar->AddChild( pcImageEditSelectAll, os::ToolBar::TB_FIXED_WIDTH );

	pcToolBar->AddSeparator("");

	// Create the redo button
	pcToolBarIcon = new os::BitmapImage();
	pcToolBarIcon->Load( cRes.GetResourceStream("editredo16x16.png") );
	pcImageEditUndo = new os::ImageButton( os::Rect(), "appwindow_redo", "Redo", new os::Message(M_BUT_EDIT_REDO), pcToolBarIcon, os::ImageButton::IB_TEXT_BOTTOM, true, true, true );
	pcToolBar->AddChild( pcImageEditUndo, os::ToolBar::TB_FIXED_WIDTH );

	// Create the undo button
	pcToolBarIcon = new os::BitmapImage();
	pcToolBarIcon->Load( cRes.GetResourceStream("editundo16x16.png") );
	pcImageEditRedo = new os::ImageButton( os::Rect(), "appwindow_undo", "Undo", new os::Message(M_BUT_EDIT_UNDO), pcToolBarIcon, os::ImageButton::IB_TEXT_BOTTOM, true, true, true );
	pcToolBar->AddChild( pcImageEditRedo, os::ToolBar::TB_FIXED_WIDTH );

	pcToolBar->AddSeparator("");

	// Create the find button
	pcToolBarIcon = new os::BitmapImage();
	pcToolBarIcon->Load( cRes.GetResourceStream("findsearch24x24.png") );
	os::Menu* pcPopMenu = new os::Menu( os::Rect(), "", os::ITEMS_IN_COLUMN );
	pcPopMenu->AddItem( "Find...", new os::Message(M_MENU_FIND_FIND) );
	pcPopMenu->AddItem( "Replace...", new os::Message(M_MENU_FIND_REPLACE) );
	pcPopMenu->AddItem( new os::MenuSeparator() );
	pcPopMenu->AddItem( "Go to Line...", new os::Message(M_MENU_FIND_GOTO) );
	pcPopMenu->SetTargetForItems(this);
	os::PopupMenu* pcFindPop = new os::PopupMenu( os::Rect(), "appwindow_find", "Find", pcPopMenu, pcToolBarIcon );
	pcToolBar->AddChild( pcFindPop, os::ToolBar::TB_FIXED_MINIMUM );

	// Attach the button bar to the window
	AddChild(pcToolBar);
}

void AppWindow::SetWindowPosition()
{
	os::Rect cWinSize = cSettings.GetRect( "WindowPos", os::Rect(101,125,700,500) );
	ResizeTo( cWinSize.Width(), cWinSize.Height() );
	MoveTo( cWinSize.left, cWinSize.top );
}

void AppWindow::Layout()
{
	static bool bHideToolbar = true;
	static bool bHideStatusbar = true;
	os::Rect cFrame = GetBounds();

	// Menu
	os::Rect cMenuFrame( cFrame.left, cFrame.top, cFrame.right, cFrame.top + pcMenuBar->GetPreferredSize(false).y );
	pcMenuBar->SetFrame(cMenuFrame);
	cFrame.top = cMenuFrame.bottom+2;

	// Toolbar
	if( pcViewToolbar->IsChecked() )
	{
		os::Rect cToolFrame( cFrame.left, cFrame.top, cFrame.right, cFrame.top + pcToolBar->GetPreferredSize(false).y );
		pcToolBar->SetFrame(cToolFrame);
		cFrame.top = cToolFrame.bottom+2;

		if( !pcToolBar->IsVisible() && !bHideToolbar )
			pcToolBar->Show();
		else
			bHideToolbar = false;
	}
	else if( pcToolBar->IsVisible() )
		pcToolBar->Hide();

	// Statusbar
	if( pcViewStatusbar->IsChecked() )
	{
		os::Rect cStatusFrame( cFrame.left, cFrame.bottom - pcStatusBar->GetPreferredSize(false).y, cFrame.right, cFrame.bottom );
		pcStatusBar->SetFrame(cStatusFrame);
		cFrame.bottom = cStatusFrame.top;

		if( !pcStatusBar->IsVisible() && !bHideStatusbar )
			pcStatusBar->Show();
		else
			bHideStatusbar = false;
	}
	else if( pcStatusBar->IsVisible() )
		pcStatusBar->Hide();

	if( bHideToolbar || bHideStatusbar )
		bHideToolbar = bHideStatusbar = false;
	
	// Current dialog
	if( pcCurrentDialog != NULL )
	{
		os::Rect cDialogFrame( cFrame.left, cFrame.bottom - pcCurrentDialog->GetPreferredSize(false).y, cFrame.right, cFrame.bottom );
		pcCurrentDialog->SetFrame(cDialogFrame);
		cFrame.bottom = cDialogFrame.top;
	}

	// Tabview
	pcEditorView->SetFrame(cFrame);
}

void AppWindow::DispatchMessage( os::Message* pcMessage, os::Handler* pcHandler )
{
	bool bPassOn = true;

	if( pcMessage->GetCode() == os::M_KEY_DOWN )	// Check for a posible shortcut key
	{
		const char* zRaw = NULL;
		int32 nQual = 0;
		if( pcMessage->FindString( "_raw_string", &zRaw ) == 0 && pcMessage->FindInt32( "_qualifiers", &nQual ) == 0 )
			bPassOn = MapKey( zRaw, nQual );
	}

	if( pcMessage->GetCode() == os::M_MOUSE_DOWN )	// Check for a mouse-click event
	{
		int32 nButtons = 0;
		os::Point cPosition;
		if( pcMessage->FindInt32( "_buttons", &nButtons ) == 0 && pcMessage->FindPoint( "_scr_pos", &cPosition ) == 0 )
			pcEditorView->MouseDown( cPosition, nButtons );
	}

	// Pass on the Message to the base class (If the MapKey() method lets us)
	if( bPassOn )
		Window::DispatchMessage( pcMessage, pcHandler );
}

bool AppWindow::MapKey( const char* pzString, int32 nQualifiers )
{
	if( nQualifiers & os::QUAL_CTRL )
	{
		if( !(nQualifiers & os::QUAL_REPEAT) )
		{
			switch( *pzString )
			{
				case 'n':	// New (Ctrl+N)
				{
					HandleMessage( new os::Message(M_MENU_FILE_NEW) );
					return(false);
				}
				case 'o':	// Open (Ctrl+O)
				{
					HandleMessage( new os::Message(M_MENU_FILE_OPEN) );
					return(false);
				}
				case 's':	// Save (Ctrl+S)
				{
					HandleMessage( new os::Message(M_MENU_FILE_SAVE) );
					return(false);
				}
				case 'x':	// Cut (Ctrl+X)
				{
					HandleMessage( new os::Message(M_MENU_EDIT_CUT) );
					return(false);
				}
				case 'c':	// Copy (Ctrl+C)
				{
					HandleMessage( new os::Message(M_MENU_EDIT_COPY) );
					return(false);
				}
				case 'v':	// Paste (Ctrl+V)
				{
					HandleMessage( new os::Message(M_MENU_EDIT_PASTE) );
					return(false);
				}
				case 'a':	// Select All (Ctrl+A)
				{
					HandleMessage( new os::Message(M_MENU_EDIT_SELECT_ALL) );
					return(false);
				}
				case 'y':	// Undo (Ctrl+Y)
				{
					HandleMessage( new os::Message(M_MENU_EDIT_REDO) );
					return(false);
				}
				case 'z':	// Undo (Ctrl+Z)
				{
					HandleMessage( new os::Message(M_MENU_EDIT_UNDO) );
					return(false);
				}
				case 'w':	// Close (Ctrl+W)
				{
					HandleMessage( new os::Message(M_MENU_FILE_CLOSE) );
					return(false);
				}
				case 'f':	// Find (Ctrl+F)
				{
					HandleMessage( new os::Message(M_MENU_FIND_FIND) );
					return(false);
				}
				case 'r':	// Replace (Ctrl+R)
				{
					HandleMessage( new os::Message(M_MENU_FIND_REPLACE) );
					return(false);
				}
				case 'g':	// Goto Line (Ctrl+G)
				{
					HandleMessage( new os::Message(M_MENU_FIND_GOTO) );
					return(false);
				}
				case 'h':	// Help Shortcuts (Ctrl+H)
				{
					HandleMessage( new os::Message(M_MENU_HELP_SHORTCUTS) );
					return(false);
				}
				case 'q':	// Quit (Ctrl+Q)
				{
					HandleMessage( new os::Message(M_MENU_APP_QUIT) );
					return(false);
				}
			}
		}
	}	
	return(true);
}

void AppWindow::HandleMessage( os::Message* pcMessage )
{
	switch( pcMessage->GetCode() )	//Get the message code from the message
	{
		case M_MENU_FILE_NEW:
		case M_BUT_FILE_NEW:
		{		
			CreateNewBuffer();
			break;
		}
		case M_MENU_FILE_OPEN:
		case M_BUT_FILE_OPEN:
		{
			// Create fileselector
			pcLoadRequester = new os::FileRequester( os::FileRequester::LOAD_REQ, new os::Messenger(this), LatestOpenPath );
			pcLoadRequester->Lock();
			pcLoadRequester->Start();
			pcLoadRequester->CenterInWindow(this);	
			if(pcLoadRequester->IsVisible())
				pcLoadRequester->Show(false);
			pcLoadRequester->Show(true);
			pcLoadRequester->MakeFocus();
			pcLoadRequester->Unlock();
			break;
		}
		case os::M_LOAD_REQUESTED:
		{
			const char* pzFPath;

			os::FileRequester *pcRequester;
			if( pcMessage->FindPointer( "source", (void**)&pcRequester ) == EOK )
				pcRequester->Close();

			if( pcMessage->FindString( "file/path",&pzFPath ) == 0 )
			{
				int n=0;

				// For each file selected in the filerequester, open it
				while( pcMessage->FindString( "file/path", &pzFPath, n ) == 0 )
				{
					Load( (char*)pzFPath );
					n++;
				}

				// Remember path
				LatestOpenPath = os::String(pzFPath);
			}		

			// Update menu items
			UpdateMenuItemStatus();
			break;
		}
		case M_MENU_FILE_CLOSE:
		case M_BUT_FILE_CLOSE:
		{
			if( pcImageFileClose->IsEnabled() )
			{
				// check if closing buffer needs to be saved first
				if( pcCurrentBuffer->NeedToSave() )
				{
					os::Alert* pcCloseAlert = new os::Alert( "Error", "Do you want to save the file before closing it?", os::Alert::ALERT_WARNING, "Save", "Discard", "Cancel", NULL );
					pcCloseAlert->CenterInWindow(this);
					pcCloseAlert->Go( new os::Invoker( new os::Message(M_INVOKED_CLOSE_ALERT), this ) );
				}
				else // If the buffer doesn't need to be save just delete the buffer
				{
					SetStatus( os::String().Format( "File %s has been closed.", pcCurrentBuffer->GetRealName().c_str() ) );
					delete(pcCurrentBuffer);
			   	}

				// Update menu items
				UpdateMenuItemStatus();
			}
			break;
		}
		case M_MENU_FILE_SAVE:
		case M_BUT_FILE_SAVE:
		{
			if( pcImageFileSave->IsEnabled() )
			{
				if(pcCurrentBuffer->Save())
				{
					SetStatus( os::String().Format( "File %s has been saved.", pcCurrentBuffer->GetRealName().c_str() ) );
				}
				else
				{
					os::String f;
					f.Format( "Unable to save file %s.", pcCurrentBuffer->GetRealName().c_str() );

					os::Alert* pcSaveErrorAlert = new os::Alert( "Error", f, os::Alert::ALERT_WARNING, "OK", NULL );
					pcSaveErrorAlert->Go( new os::Invoker() );
					SetStatus(f);	
				}

				// Update menu items
				UpdateMenuItemStatus();
			}
			break;
		}
		case M_MENU_FILE_SAVE_AS:
		case M_BUT_FILE_SAVE_AS:
		{
			if( pcImageFileSaveAs->IsEnabled() )
			{
				pcSaveRequester = new os::FileRequester( os::FileRequester::SAVE_REQ, new os::Messenger(this), LatestSavePath );
				pcSaveRequester->Lock();
			   	pcSaveRequester->Start();
				pcSaveRequester->CenterInWindow(this);
				if(pcSaveRequester->IsVisible())
					pcSaveRequester->Show(false);
 				pcSaveRequester->Show(true);
				pcSaveRequester->MakeFocus();
				pcSaveRequester->Unlock();
			}
			break;
		}
		case os::M_SAVE_REQUESTED:
		{
			const char* pzFPath;

			os::FileRequester *pcRequester;
			if( pcMessage->FindPointer( "source", (void**)&pcRequester ) == EOK )
				pcRequester->Close();

			if( pcMessage->FindString( "file/path", &pzFPath ) == 0 )
			{
				if( pcCurrentBuffer != NULL )
				{
					if( pcCurrentBuffer->SaveAs(pzFPath) )
					{
						SetStatus( os::String().Format( "File %s has been saved.", pcCurrentBuffer->GetRealName().c_str() ) );
					}
					else
					{
						os::String f;
						f.Format( "Unable to save file %s.", pcCurrentBuffer->GetRealName().c_str() );

						os::Alert* pcSaveErrorAlert = new os::Alert( "Error", f, os::Alert::ALERT_WARNING, "OK", NULL );
						pcSaveErrorAlert->Go( new os::Invoker() );
						SetStatus(f);		
					}
				}

				// Update menu items
				UpdateMenuItemStatus();

				// Remember path
				LatestSavePath = os::String(pzFPath);
			}
			break;
		}
		case M_MENU_EDIT_CUT:
		case M_BUT_EDIT_CUT:
		{
			if( pcEditCut->IsEnabled() )
				pcCurrentBuffer->Cut();
			break;
		}
		case M_MENU_EDIT_COPY:
		case M_BUT_EDIT_COPY:
		{
			if( pcEditCopy->IsEnabled() )
				pcCurrentBuffer->Copy();
			break;
		}
		case M_MENU_EDIT_PASTE:
		case M_BUT_EDIT_PASTE:
		{
			if( pcEditPaste->IsEnabled() )
				pcCurrentBuffer->Paste();
			break;
		}
		case M_MENU_EDIT_SELECT_ALL:
		case M_BUT_EDIT_SELECT_ALL:
		{
			if( pcEditSelectAll->IsEnabled() )
				pcCurrentBuffer->SelectAll();
			break;
		}

		case M_MENU_EDIT_UNDO:
		case M_BUT_EDIT_UNDO:
		{
			if( pcEditUndo->IsEnabled() )
				pcCurrentBuffer->Undo();
			break;
		}
		case M_MENU_EDIT_REDO:
		case M_BUT_EDIT_REDO:
		{
			if( pcEditRedo->IsEnabled() )
				pcCurrentBuffer->Redo();
			break;
		}
		case M_MENU_VIEW_TOOLBAR:
			Layout();
			break;
		case M_MENU_VIEW_STATUSBAR:
			Layout();
			break;
		case M_MENU_VIEW_SETFONT:
		{
			if( pcFontRequester == NULL )
				pcFontRequester = new os::FontRequester( new os::Messenger(this), true);

			pcFontRequester->CenterInWindow(this);
			if( pcFontRequester->IsVisible() )
				pcFontRequester->Show(false);
			pcFontRequester->Show();
			pcFontRequester->MakeFocus();
			break;
		}
		case os::M_FONT_REQUESTED:
		{
			os::Font* f = new os::Font();
			if( pcMessage->FindObject( "font", *f ) == EOK )
			{
				if( pcCurrentBuffer != NULL )
					pcCurrentBuffer->SetFont(f);
			}
			break;
		}
		case M_MENU_FIND_FIND:
		case M_BUT_FIND_FIND:
		{
			// Create dialog if it doesn't exists
			if( pcFindDialog == NULL )
				pcFindDialog = new FindDialog();

			UpdateMenuItemStatus();
			SetDialog(pcFindDialog);
			break;
		}
		case M_MENU_FIND_REPLACE:
		{
			// Create dialog if it doesn't exists
			if( pcReplaceDialog == NULL )
				pcReplaceDialog = new ReplaceDialog();

			UpdateMenuItemStatus();
 			SetDialog(pcReplaceDialog);
			break;
		}
		case M_MENU_FIND_GOTO:
		{
			// Create dialog if it doesn't exists
			if( pcGotoDialog == NULL )
				pcGotoDialog = new GotoDialog();

			UpdateMenuItemStatus();
			SetDialog(pcGotoDialog);
			break;
		}
		case M_MENU_SETTINGS_SAVE_ON_CLOSE:
		{
			break;
		}
		case M_MENU_SETTINGS_SAVE_NOW:
		{
			cSettings.SetBool( "SaveOnExit", ( (os::CheckMenu*)pcSettingsSaveOnClose )->IsChecked() );
			cSettings.SetRect( "WindowPos", GetFrame() );
			cSettings.Save();
			break;
		}
		case M_MENU_SETTINGS_RESET:
		{
			os::Rect cWinSize( 100, 125, 100+600, 125+375 );
			ResizeTo( cWinSize.Width(), cWinSize.Height() );
			MoveTo( cWinSize.left, cWinSize.top );
			cSettings.SetRect( "SetWindowPos", cWinSize );
			((os::CheckMenu*)pcSettingsSaveOnClose)->SetChecked(true);			
			LatestOpenPath = "";
			LatestSavePath = "";
			break;
		}
		case M_MENU_HELP_SHORTCUTS:
		{
			os::String HelpTitle = "Keyboard Shortcuts...";
			os::String AppShortcuts = "App Quit - Ctrl+Q\n\n";
			os::String FileShortcuts = "File New - Ctrl+N\nFile Open - Ctrl+O\nFile Save - Ctrl+S\nFile Save As -\nFile Close - Ctrl+W\n\n";
			os::String EditShortcuts = "Edit Cut - Ctrl+X\nEdit Copy - Ctrl+C\nEdit Paste - Ctrl+V\nEdit Select All - Ctrl+A\nEdit Redo - Ctrl+Y\nEdit Undo - Ctrl+Z\n\n";
			os::String FindShortcuts = "Find Find - Ctrl+F\nFind Replace - Ctrl+R\nFind Goto - Ctrl+G";

			os::Alert* pcHelp = new os::Alert( HelpTitle, AppShortcuts+FileShortcuts+EditShortcuts+FindShortcuts, os::Alert::ALERT_INFO, 0x00, "Close", NULL );
			pcHelp->CenterInWindow(this);
			pcHelp->Go(new os::Invoker());
			if(pcHelp->IsVisible())
				pcHelp->Show(false);
			pcHelp->Show(true);
			pcHelp->MakeFocus();
			break;
		}
		case M_INVOKED_EDITOR:
		{
			int32 nEvent;

			if(pcMessage->FindInt32( "events", &nEvent) == 0 )
			{
				if( pcCurrentBuffer != NULL )
					pcCurrentBuffer->Invoked(nEvent);
			}
			break;
		}
		case M_INVOKED_TAB_CHANGED:  // User has changed the tab. Update pointer to current buffer
		{
			// Get new tab
			int iSel = pcEditorView->GetSelection();
			
			// Make this page the current one
			SetCurrentTab(iSel);

			// Set cursor position
			pcCurrentBuffer->Invoked( os::TextView::EI_CURSOR_MOVED );

			// Set focus on this buffer
			pcCurrentBuffer->MakeFocus();

			// Update menu items
			UpdateMenuItemStatus();

			break;
		}
		case M_INVOKED_CLOSE_ALERT:   // The alert created in M_MENU_FILE_CLOSE
		{
			int32 nSelection;

			if( pcMessage->FindInt32( "which", &nSelection ) )
				break;

			switch( nSelection )	
			{
				case 0:  // User wants to save it
				{
					// Trying to save
					if( !pcCurrentBuffer->Save() )
					{
						pcSaveAsRequester = new os::FileRequester( os::FileRequester::SAVE_REQ, new os::Messenger(this), LatestSavePath, 
											os::FileRequester::NODE_FILE, false, new os::Message( M_INVOKED_SAVEAS_ALERT ), NULL, false );
						pcSaveAsRequester->Lock();
						pcSaveAsRequester->Start();						
						if(pcSaveAsRequester->IsVisible())
							pcSaveAsRequester->Show(false);
						pcSaveAsRequester->Show(true);
						pcSaveAsRequester->MakeFocus();
						pcSaveAsRequester->Unlock();
					}
					else
					{
						// Delete the buffer (in the destructor the new current buffer will be set)
						SetStatus( os::String().Format( "File %s has been closed and changes has been saved.", pcCurrentBuffer->GetRealName().c_str() ) );

						// Update menu items
						UpdateMenuItemStatus();
						delete(pcCurrentBuffer);					
					}
					break;
				}
				case 1:  // User doesn't want to save it
				{					
					SetStatus( os::String().Format( "File %s has been closed without saving changes.", pcCurrentBuffer->GetRealName().c_str() ) );

					// Delete the buffer (in the destructor the new current buffer will be set)
					delete(pcCurrentBuffer);
					// Update menu items
					UpdateMenuItemStatus();
					break;
				}
				default: // User cancels the operation
					break;
			}
			break;
		}
		case M_INVOKED_SAVEAS_ALERT:   // Created in M_INVOKED_CLOSE_ALERT by pcSaveAsRequester
		{
			const char* pzFPath;

			if( pcMessage->FindString( "file/path", &pzFPath ) == 0 )
			{
				if( pcCurrentBuffer != NULL )
				{
					if( pcCurrentBuffer->SaveAs(pzFPath) )
					{
						SetStatus( os::String().Format( "File %s has been closed and changes has been saved.", pcCurrentBuffer->GetRealName().c_str() ) );
						delete(pcCurrentBuffer);

						// Update menu items
						UpdateMenuItemStatus();
					}
					else
					{
						os::String f;
						f.Format( "Unable to save file %s.", pcCurrentBuffer->GetRealName().c_str() );

						os::Alert* pcSaveErrorAlert = new os::Alert( "Error", f, os::Alert::ALERT_WARNING, "OK", NULL );
						pcSaveErrorAlert->Go( new os::Invoker() );
						SetStatus(f);		
					}
				}
			}
			pcSaveAsRequester = NULL;
			break;
		}
		case M_INVOKED_QUIT_ALERT: // Check if the user is sure to quit
		{
			int32 nSelection;

			if( pcMessage->FindInt32( "which", &nSelection ) )
				break;
					
			if( nSelection == 0 )	
			{
				if( ( (os::CheckMenu*)pcSettingsSaveOnClose )->IsChecked() )
				{
					cSettings.SetBool( "SaveOnExit", true );
					cSettings.SetRect( "WindowPos", GetFrame() );
					cSettings.SetString( "LoadPath", LatestOpenPath );
					cSettings.SetString( "SavePath", LatestSavePath );
					cSettings.Save();
				}			
				os::Application::GetInstance()->PostMessage( os::M_QUIT );
			}
			break;
		}
		case M_MENU_APP_ABOUT:
		{
			os::String sAbout;
			sAbout = APP_VERSION_STRING + os::String("\n\n");
			sAbout += APP_DESC_STRING + os::String("\n");
			sAbout += APP_DATE_STRING + os::String("\n\n");
			sAbout += APP_CREDIT_STRING + os::String("\n\n");
			sAbout += APP_GPL_STRING;

			os::Alert* pcAbout	= new os::Alert( APP_LABEL_STRING, sAbout, os::Alert::ALERT_INFO, 0x00, "Close", NULL );
			pcAbout->CenterInWindow(this);
			pcAbout->Go(new os::Invoker());
			if(pcAbout->IsVisible())
				pcAbout->Show(false);
			pcAbout->Show(true);
			pcAbout->MakeFocus();
			break;
		}
		case M_MENU_APP_QUIT:
		{
			OkToQuit();
			break;
		}
		default:
		{
			Window::HandleMessage(pcMessage);
			break;
		}
	}
}

void AppWindow::UpdateMenuItemStatus()
{
	pcFileClose->SetEnable( pcCurrentBuffer != NULL );
	pcFileSave->SetEnable( pcCurrentBuffer != NULL );
	pcFileSaveAs->SetEnable( pcCurrentBuffer !=NULL );
	pcEditUndo->SetEnable( pcCurrentBuffer != NULL );
	pcEditRedo->SetEnable( pcCurrentBuffer != NULL );
	pcEditCut->SetEnable( pcCurrentBuffer != NULL );
	pcEditCopy->SetEnable( pcCurrentBuffer != NULL );
	pcEditPaste->SetEnable( pcCurrentBuffer != NULL );
	pcEditSelectAll->SetEnable( pcCurrentBuffer != NULL );

	// Check if the menuitem "save" shall be enabled. It should only be
    // enable if the buffer has been modified since last saved
	if( pcCurrentBuffer != NULL && !pcCurrentBuffer->SaveStatus() )
		pcFileSave->SetEnable(false);

	// Set also enable/disable for the find, replace and goto dialogs
	if( pcFindDialog != NULL )
		pcFindDialog->SetEnable( pcCurrentBuffer != NULL );	
	if( pcReplaceDialog!= NULL )
		pcReplaceDialog->SetEnable( pcCurrentBuffer != NULL );	
	if( pcGotoDialog != NULL )
		pcGotoDialog->SetEnable( pcCurrentBuffer != NULL );	

	// Check button bar
	pcImageFileClose->SetEnable( pcFileClose->IsEnabled() );
	pcImageFileSave->SetEnable( pcFileSave->IsEnabled() );
	pcImageFileSaveAs->SetEnable( pcFileSaveAs->IsEnabled() );
	pcImageEditCopy->SetEnable( pcEditCopy->IsEnabled() );
	pcImageEditCut->SetEnable( pcEditCut->IsEnabled() );
	pcImageEditPaste->SetEnable( pcEditPaste->IsEnabled() );
	pcImageEditSelectAll->SetEnable( pcEditSelectAll->IsEnabled() );
	pcImageEditUndo->SetEnable( pcEditUndo->IsEnabled() );
	pcImageEditRedo->SetEnable( pcEditRedo->IsEnabled() );
}

void AppWindow::CreateNewBuffer()
{
	// Create the new buffer
	Buffer *buf = new Buffer( this, NULL ,NULL );

	// Attach new buffer to tabview
	int iSel = pcEditorView->AppendTab( buf->GetTabNameOfBuffer(), buf->GetEditView() );

	// Make this page the current one
	pcEditorView->SetSelection( iSel, false );

	// Set status bar
	SetStatus( os::String().Format( "New file %s has been created.", pcCurrentBuffer->GetRealName().c_str() ) );

	// Update menu items
	UpdateMenuItemStatus();

	// Force a complete redraw
	buf->ForceRedraw();		

	// And set focus
	buf->MakeFocus();
}

void AppWindow::DragAndDrop( os::Message* pcData )
{
	int i = 0;
	const char* pzFPath;

	while( pcData->FindString( "file/path", &pzFPath, i ) == 0 )
	{
		Load( (char*)pzFPath );
		i++;
	}
}

void AppWindow::Load( char* pzFileName )
{
	uint32 nFileSize = 0;
	struct stat sFileStat;

	// Check if this file is already opened
	int r = CheckIfAlreadyOpened( os::String(pzFileName) );

	if( r != -1 )
	{
		// Set the tab	
		pcEditorView->SetSelection(r);
		return;
	}

	// Load it
	if( !stat( pzFileName, &sFileStat ) )
	{
		if( sFileStat.st_mode & S_IFREG )
		{
			// File is OK, we can get the size & load it
			nFileSize = sFileStat.st_size;

			char* pnBuffer = (char*)malloc( nFileSize + 1 );
			pnBuffer = (char*)memset( pnBuffer, 0, nFileSize + 1 );

			uint32 nCount, nCount2 = 0;

			FILE *hFile = fopen( pzFileName, "r" );
			if( hFile )
			{
				for( nCount = 0, nCount2 = 0; nCount < nFileSize; nCount++)
				{
					char c = fgetc(hFile);
					if( c != '\r' )
					{
						pnBuffer[nCount2] = c;
						nCount2++;
					}
				}
			}
			fclose(hFile);
			pnBuffer[nCount2] = 0;  // Should have been set by memset, but just in case

			// File has been loaded. Let's open a new Buffer for this file

			// Create the new buffer
			Buffer *buf = new Buffer( this, pzFileName, pnBuffer );

			// Attach new buffer to tabview
			int iSel = pcEditorView->AppendTab( buf->GetTabNameOfBuffer(), buf->GetEditView() );

			// Make this page the current one. Don't send a message as it will screw up the status message
			pcEditorView->SetSelection( iSel, false );

			// since the file has been loaded we can set the ContentChanged flag to false
			buf->ContentChanged(false);

			// Set status bar
			SetStatus( os::String().Format( "File %s has been successfully opened.", pcCurrentBuffer->GetRealName().c_str() ) );

			// Update menu items
			UpdateMenuItemStatus();

			// Make focus
			buf->MakeFocus();

			buf->ForceRedraw();

			free(pnBuffer);
		}
		else
		{
			// Set status bar
			os::String f;
			f.Format( "Unable to open file %s.", pzFileName );
			SetStatus(f);

			// Pop error window, file is not a "regular" file
			os::Alert* pcFileAlert = new os::Alert( "Error", f, os::Alert::ALERT_WARNING, 0x00, "OK", NULL );
			pcFileAlert->Go();
		}
	}
	else  // File didn't exists. Open it anyway
	{
		// Create the new buffer
		Buffer *buf = new Buffer( this, pzFileName, NULL );

		// Attach new buffer to tabview
		int iSel = pcEditorView->AppendTab( buf->GetTabNameOfBuffer(), buf->GetEditView() );

		// Make this page the current one
		pcEditorView->SetSelection( iSel, false );

		// since the file has been loaded we can set the ContentChanged flag to false
		buf->ContentChanged(false);

		// Set status bar
		SetStatus( os::String().Format( "New file %s has been created.", pzFileName ) );

		// Update menu items
		UpdateMenuItemStatus();

		// Make focus
		buf->MakeFocus();	
	}
}

// Set the status bar
void AppWindow::SetStatus( os::String zTitle, ... )
{
	os::String f;
	va_list sArgList;

	va_start( sArgList, zTitle.c_str() );
	f.Format( zTitle.c_str(), sArgList );
	va_end( sArgList );

	pcStatusBar->SetText( "message", f );
}

void AppWindow::SetDialog( Dialog* dialog )
{
	if( pcCurrentDialog != NULL )
	{
		pcCurrentDialog->Close();
		RemoveChild(pcCurrentDialog);
		pcCurrentDialog	= NULL;
	}

	if( dialog != NULL )
	{
		pcCurrentDialog = dialog;
		AddChild(pcCurrentDialog);
		pcCurrentDialog->Init();
	}

	Layout();
}

bool AppWindow::OkToQuit()
{
	if( AnyBufferNeedsToBeSaved() ) // If yes display nice alert
	{
		os::String AlertTitle = "Really quit";
		os::String AlertMessage = "You have unsaved files. Data will be lost if you quit.\nAre you sure you want to quit?";
		os::Alert* pcQuitAlert = new os::Alert( AlertTitle, AlertMessage, os::Alert::ALERT_QUESTION, 0x00, "Quit", "Cancel", NULL );
		pcQuitAlert->CenterInWindow(this);
		pcQuitAlert->Go( new os::Invoker( new os::Message( M_INVOKED_QUIT_ALERT ), this ) );
	}
	else  // if not send a message that application can be closed 
	{     // (same message as if the user press "yes" on the above alert)
		os::Message *msg = new os::Message( M_INVOKED_QUIT_ALERT );
		msg->AddInt32( "which", 0 );
		PostMessage( msg, this );
	}
	return(false);
}
