/*  CodeEdit -:-  (C)opyright 2001-2002 Andreas Engh-Halstvedt
 *  AEdit -:-     (C)opyright 2000-2002 Kristian Van Der Vliet
 *  AEdit -:-     (C)opyright 2004-2006 Jonas Jarvoll
 *  CodeEdit -:-  (C)opyright 2011-2012 David Kent
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details. 
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <iostream>
#include <string>

#include <buffer.h>
#include <messages.h>
#include <editview.h>
#include <version.h>

#include <codeview/format_c.h>


// Declare some static variables
int Buffer::iEmptyBufferCounter = 0;
std::list< Buffer* > Buffer::pcBufferList;
AppWindow* Buffer::pcTarget = NULL;

Buffer::Buffer( AppWindow* pcTar, char* pFileName, char* pFileBuffer )
{
	// One more new buffer if no specified filename
	if( pFileName == NULL ) 
		iEmptyBufferCounter++;

	// Set target
	pcTarget = pcTar;

	// Create EditView for this buffer
	pcEditView = new EditView( os::Rect(0,0,0,0), pcTar );
	pcEditView->SetTarget( pcTarget );

	// Set buffer if we have a buffer
	if( pFileBuffer != NULL )
	{
		pcEditView->Clear(true);		
		pcEditView->Set( pFileBuffer, true );
	}

	pcEditView->SetMessage( new os::Message(M_INVOKED_EDITOR) );
	pcEditView->SetEventMask( cv::CodeView::EI_CONTENT_CHANGED | cv::CodeView::EI_CURSOR_MOVED );
	pcEditView->SetShowLineNumbers(true);

	// Set CodeView colors
	pcEditView->SetHighlightColor( os::Color32_s(255,235,185) );
	pcEditView->SetBackground( os::Color32_s(255,255,255) );
	pcEditView->SetTextColor( os::Color32_s(0,0,0) );
	pcEditView->SetLineNumberBgColor( os::Color32_s(220,220,220) );
	pcEditView->SetFormat( new cv::Format_C() );

    // Get Default defined font
	os::font_properties fp;
	os::Font::GetDefaultFont( DEFAULT_FONT_FIXED, &fp );
	os::Font *f = new os::Font(fp);
	pcEditView->SetFont(f);

	// Set up variables
	vSize = fp.m_vSize;
	zFamily = fp.m_cFamily;
	zStyle = fp.m_cStyle;
	iFlags = fp.m_nFlags;

	f->Release();

	// Make sure this EditView has the focus
	pcEditView->MakeFocus();

	// Set filename
	if( pFileName == NULL )
		zFilename = "";
	else
		zFilename = pFileName;

	// Reset the change flag
	bContentChanged = false;

	// No find has been done
	bHasFind = false;

	// Since this is a new buffer we need to remember the EmptyBufferNumber
	// This number is used to create a nice name of the buffer
	iEmptyBufferNumber = iEmptyBufferCounter;

	// Add this buffer to list of buffers
	pcBufferList.push_back(this);

	// Make this the current one
	MakeCurrent();
}


Buffer::~Buffer()
{
	std::list< Buffer* >::iterator iter;
	int icounter;

	// Is this buffer the current one?
	if( pcTarget->pcCurrentBuffer == this )
		pcTarget->pcCurrentBuffer = NULL;

	// Remove buffer from list
	for( iter = pcBufferList.begin(), icounter = 0; iter != pcBufferList.end(); iter++, icounter++ )
	{
		if( *iter == this )
		{
			// remove tab
			pcTarget->pcEditorView->DeleteTab(icounter);
			// remove buffer from list
			pcBufferList.erase(iter);
			break;
		}
	}
	
	// Set current buffer
	if( pcTarget->pcEditorView->GetTabCount() > 0 )
		SetCurrentTab( pcTarget->pcEditorView->GetSelection() );
	else  // No more open document
	{
		// Set the window title etc.
		pcTarget->SetTitle(APP_VERSION_STRING);
	}
	
	delete(pcEditView);
}

// Get real name of buffer (without a asterix)
os::String &Buffer::GetRealName()
{
	static os::String zTitle2;
	static char zTitle[32];

	// Check if a filename already exists
	if( zFilename != "" )
		return(zFilename);

	// Otherwise we create one (this must be done in a nicer way how??)
	sprintf( zTitle, "Untitled - %d", iEmptyBufferNumber );
	zTitle2 = zTitle;

	return(zTitle2);
}

// Get the name of the buffer
// Normal cases it will return the path to the file
// if no file exits it will return "Noname"
os::String &Buffer::GetNameOfBuffer()
{
	static os::String str;

	if( bContentChanged )
	{
		str = "*";
		str += GetRealName();
		return(str);
	}
	else
		return GetRealName();
}

// Gets the filename of the buffer
// Will delete the path from the name
std::string &Buffer::GetTabNameOfBuffer()
{
	std::string str;
	static std::string ret;
	unsigned int i;

	str = GetRealName();
	i = str.find_last_of( "/" );

	// clear string
	ret.assign( "" );

	// Add a star if the buffer has bene modifed
	if( bContentChanged )
		ret.assign( "*" );

	// Append the name of the buffer
	if( i == std::string::npos )
		ret.append(str);
	else
		ret.append( str, i+1, str.size() );

	return(ret);
}


// Make this buffer the current one
void Buffer::MakeCurrent()
{
	pcTarget->pcCurrentBuffer = this;

	// Change appwindow title
	os::String zWindowTitle = APP_VERSION_STRING;
	zWindowTitle += " : ";
	zWindowTitle += GetNameOfBuffer();
	pcTarget->SetTitle( zWindowTitle );

	// Since we have changed the current buffer the find is no longer valid
	bHasFind = false;
}

// Content or cursor changed
void Buffer::Invoked( int32 nEvent )
{
	// The content in the cv::CodeView has been changed
	if( !bContentChanged && (nEvent & cv::CodeView::EI_CONTENT_CHANGED) )
		ContentChanged( true );

	// Update cursor location
	if( nEvent & cv::CodeView::EI_CURSOR_MOVED )
	{
		pcEditView->GetCursor();
	}
}

// Returns true if buffer has ben modified and there is a proper filename
// for the buffer
bool Buffer::SaveStatus( void )
{
	return( bContentChanged && zFilename != "" );
}

// Content has changed
void Buffer::ContentChanged( bool iContent )
{
	std::list< Buffer* >::iterator iter;
	int icounter;

	bContentChanged = iContent;
	
	// Change name of tab by go through list to find matching buffer
	for( iter = pcTarget->pcCurrentBuffer->pcBufferList.begin(), icounter = 0; iter != pcTarget->pcCurrentBuffer->pcBufferList.end(); iter++, icounter++ )
	{
		if( *iter == this )
			pcTarget->pcEditorView->SetTabTitle( icounter, GetTabNameOfBuffer() );
	}
	
	// Change name of window if it is the current one
	if( pcTarget->pcCurrentBuffer == this )
		MakeCurrent();

	// Also update the "Save" menu
	pcTarget->UpdateMenuItemStatus();
}

// Tries to save the buffer. Returns FALSE is anything wrong
bool Buffer::Save()
{
	return( SaveAs(zFilename.c_str()) );
}

bool Buffer::SaveAs( const char* pFileName )
{
	bool bSuccess = false;

	try
	{
		os::File cFile( pFileName, O_TRUNC | O_CREAT | O_WRONLY );
		os::String cLine = "";

		uint32 nLines = pcEditView->GetLineCount();
		for( uint32 nLineCount = 0; nLineCount <= nLines; nLineCount++ )
		{
			if( nLineCount > 0 )
			{
				cLine = os::String("\n") + pcEditView->GetLine(nLineCount);
			}
			else
			{
				cLine = pcEditView->GetLine(nLineCount);
			}
			cFile.Write( cLine.c_str(), cLine.size() );
		}
		cFile.Flush();

		// Change the name of the buffer
		zFilename = pFileName;

		ContentChanged(false);
		bSuccess = true;
	}
	catch( os::errno_exception &e )
	{
		std::cerr << "Error saving \"" << pFileName << "\": %s" << e.what() << std::endl;
		bSuccess = false;
	}

	return bSuccess;
}

// Move the cursor to a specified line number
void Buffer::GotoLine( int32 iLineNo )
{
	pcEditView->SetCursor( 0, (iLineNo - 1), false, false );
}

// Make sure cv::CodeView has the focus
void Buffer::MakeFocus()
{
	pcEditView->MakeFocus();
}

void Buffer::SetFont( os::Font* font )
{
	pcEditView->SetFont(font);
}

// Changes the font in the cv::CodeView
void Buffer::SetFontFamily( const char* cFam, const char* cStyle, const uint32 iF )
{
	os::Font *font = new os::Font();

	font->SetFamilyAndStyle( cFam, cStyle );
	font->SetSize(vSize);

	pcEditView->SetFont(font);

	zFamily = cFam;
	zStyle = cStyle;
	iFlags = iF;
	
	font->Release();
}

void Buffer::SetFontSize( float size )
{
	os::Font *font = new os::Font();

	font->SetFamilyAndStyle( zFamily.c_str(), zStyle.c_str() );
	font->SetSize(size);
	font->SetFlags( iFlags);

	pcEditView->SetFont(font);

	vSize = size;
	
	font->Release();
}

bool Buffer::FindFirst( os::String zText, bool bCaseSens )
{
	// Setup buffer variables
	bHasFind = false;
	zFindText = zText;
	bCaseSensitive = bCaseSens;

	nTotalLines = pcEditView->GetLineCount();

	for( nCurrentLine = 0; nCurrentLine < nTotalLines; nCurrentLine++ )
	{
		nTextPosition = pcEditView->Find( pcEditView->GetLine( nCurrentLine ), zFindText, bCaseSensitive, 0 );

		if( nTextPosition != -1 )	// Find() returns -1 if the text isn't found
		{
			pcEditView->SetCursor( nTextPosition, nCurrentLine, false, false );
			pcEditView->SetCursor( (nTextPosition + zFindText.size()), nCurrentLine, true, false );
			bHasFind = true;
			break;
		}
	}

	return(bHasFind);
}

os::String &Buffer::FindNext()
{
	static os::String sText = "";

	if( bHasFind )
	{
		nTextPosition += zFindText.size();	// We want to start the search from just past the last instance.
					 						// There is a lot of fudging in this block of code :)

		for( ; nCurrentLine < nTotalLines; nCurrentLine++ )
		{
			nTextPosition = pcEditView->Find( pcEditView->GetLine( nCurrentLine ), zFindText, bCaseSensitive, nTextPosition );

			// We have found something here!
			if( nTextPosition != -1 )
			{
				pcEditView->SetCursor( nTextPosition, nCurrentLine, false, false );
				pcEditView->SetCursor( (nTextPosition + zFindText.size()), nCurrentLine, true, false );
				return(zFindText);
			}
			else
			{
				nTextPosition = 0;		// This is a big fudge (See the call to EditText::Find() above!)
			}
		}
	}

	return(sText);
}

bool Buffer::Replace( os::String zFText, os::String zRText )
{
	zFindText = zFText;
	zReplaceText = zRText;

	if( bHasFind )
	{
		pcEditView->Delete();
		pcEditView->Insert( zReplaceText.c_str(), false );

		return(true);
	}

	return(false);
}

// Forces a redraw of the EditView
void Buffer::ForceRedraw()
{
	pcEditView->Invalidate(true);
	pcEditView->Sync();
}

///////////////////////////////////////////////////////////////////
// Friends: 
///////////////////////////////////////////////////////////////////
void SetCurrentTab( int iSel )
{
	std::list< Buffer* >::iterator iter;
	int icounter;

	// If we do not have any more buffers

	// Go throught list to find matching buffer
	for( iter = Buffer::pcTarget->pcCurrentBuffer->pcBufferList.begin(), icounter = 0; iter != Buffer::pcTarget->pcCurrentBuffer->pcBufferList.end(); iter++, icounter++ )
	{
		if( icounter == iSel )
		{
			(*iter)->MakeCurrent();
			return;	
		}
	}
}

// Returns true if any buffer has modified context and therefor needs to be saved
bool AnyBufferNeedsToBeSaved()
{
	std::list< Buffer* >::iterator iter;

	// Go throught list to find matching buffer
	for( iter=Buffer::pcTarget->pcCurrentBuffer->pcBufferList.begin(); iter != Buffer::pcTarget->pcCurrentBuffer->pcBufferList.end(); iter++ )
	{
		if( (*iter)->NeedToSave() )
			return(true);
	}

	return(false);
}

// Returns true if Save All menu item should be active
// true - if there is modified buffers and those has a real filename (ie. not "untitled");
bool MenuItemSaveAllShallBeActive()
{
	std::list< Buffer* >::iterator iter;

	// Go throught list to find matching buffer
	for( iter = Buffer::pcTarget->pcCurrentBuffer->pcBufferList.begin(); iter != Buffer::pcTarget->pcCurrentBuffer->pcBufferList.end(); iter++ )
	{
		if( (*iter)->SaveStatus() )
			return(true);
	}

	return(false);
}

// Save all buffers which has been modified and have a real filename
// returns true if successful
bool SaveAllBuffers()
{
	bool ret = true;
	std::list< Buffer* >::iterator iter;

	// Go throught list to find matching buffer
	for( iter = Buffer::pcTarget->pcCurrentBuffer->pcBufferList.begin(); iter != Buffer::pcTarget->pcCurrentBuffer->pcBufferList.end(); iter++ )
	{
		if( !( (*iter)->Save() ) )
			ret = false;
	}

	return(ret);
}

// Returns the buffer with the name
int CheckIfAlreadyOpened( os::String name )
{
	int rv = 0;

	std::list< Buffer* >::iterator iter;

	// Go throught list to find matching buffer
	for( iter = Buffer::pcTarget->pcCurrentBuffer->pcBufferList.begin(); iter != Buffer::pcTarget->pcCurrentBuffer->pcBufferList.end(); iter++, rv++ )
	{
		if( (*iter)->GetRealName() == name )
			return(rv);
	}

	// Buffer was not found
	return(-1);
}
