/*  CodeEdit -:-  (C)opyright 2001-2002 Andreas Engh-Halstvedt
 *  AEdit -:-     (C)opyright 2000-2002 Kristian Van Der Vliet
 *  AEdit -:-     (C)opyright 2004-2006 Jonas Jarvoll
 *  CodeEdit -:-  (C)opyright 2011-2012 David Kent
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details. 
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <util/message.h>
#include <gui/textview.h>

#include <string>

#include <editview.h>
#include <messages.h>


EditView::EditView( const os::Rect& cFrame, AppWindow *pcMain ) : cv::CodeView( cFrame, "edit_view", "", os::CF_FOLLOW_BOTTOM | os::CF_FOLLOW_TOP |  os::CF_FOLLOW_LEFT | os::CF_FOLLOW_RIGHT )
{
	pcMainWindow = pcMain;

	// Create a context menu
	pcContextMenu = new os::Menu( os::Rect(0,0,10,10), "", os::ITEMS_IN_COLUMN );
	pcContextMenu->AddItem( "Cut", new os::Message(M_MENU_EDIT_CUT) );
	pcContextMenu->AddItem( "Copy", new os::Message(M_MENU_EDIT_COPY) );
	pcContextMenu->AddItem( "Paste", new os::Message(M_MENU_EDIT_PASTE) );
	pcContextMenu->AddItem( new os::MenuSeparator() );
	pcContextMenu->AddItem( "Select All", new os::Message(M_MENU_EDIT_SELECT_ALL) );
	pcContextMenu->AddItem( new os::MenuSeparator() );
	pcContextMenu->AddItem( "Undo", new os::Message(M_MENU_EDIT_UNDO) );
	pcContextMenu->AddItem( "Redo", new os::Message(M_MENU_EDIT_REDO) );
}

void EditView::AttachedToWindow()
{
	pcContextMenu->SetTargetForItems( GetWindow() );
}

void EditView::MouseDown( const os::Point &cPosition, uint32 nButtons )
{
	if( nButtons == 2 )
		pcContextMenu->Open( ConvertToScreen(cPosition) );	// Open the menu where the mouse is
	else
		cv::CodeView::MouseDown( cPosition, nButtons );
}

// We implement this function to be able to do drag and drop of files
void EditView::MouseUp( const os::Point& cPosition, uint32 nButtons, os::Message* pcData )
{
	if( pcData != NULL )
		pcMainWindow->DragAndDrop(pcData);
	else
		cv::CodeView::MouseUp( cPosition, nButtons, pcData );
}

void EditView::KeyDown( const char* pzString, const char* pzRawString, uint32 nQualifiers )
{
// This is not a nice way of doing shortcuts. Best would be to first check if the shortcut belongs to a menuitem
// and if not pass the shortcut to cv::CodeView. Unfortunetly there is no(?) way of knowing (except going through
// all menuitems) if the shortcut did have a menuitem attached to it or not.
// The main probem with the below solution is that the shortcut is defined in two places, this place but
// also where you create the menuitem.

	if( (nQualifiers & os::QUAL_CTRL) && !(nQualifiers & os::QUAL_SHIFT) && !(nQualifiers & os::QUAL_ALT) )
	{
		switch( *pzRawString )
		{
			case 'n':	// New (Ctrl+N)
			{
				if( !(nQualifiers & os::QUAL_REPEAT) )
					GetWindow()->HandleMessage( new os::Message(M_MENU_FILE_NEW) );
				return;
			}
			case 'o':	// Open (Ctrl+O)
			{
				if( !(nQualifiers & os::QUAL_REPEAT) )
					GetWindow()->HandleMessage( new os::Message(M_MENU_FILE_OPEN) );
				return;
			}
			case 's':	// Save (Ctrl+S)
			{
				if( !(nQualifiers & os::QUAL_REPEAT) )
					GetWindow()->HandleMessage( new os::Message(M_MENU_FILE_SAVE) );
				return;
			}
			case 'w':	// Close (Ctrl+W)
			{
				if( !(nQualifiers & os::QUAL_REPEAT) )
					GetWindow()->HandleMessage( new os::Message(M_MENU_FILE_CLOSE) );
				return;
			}
			case 'a':	// Select all (Ctrl+A)
			{
				if( !(nQualifiers & os::QUAL_REPEAT) )
					GetWindow()->HandleMessage( new os::Message(M_MENU_EDIT_SELECT_ALL) );
				return;
			}
			case 'f':	// Find (Ctrl+F)
			{
				if( !(nQualifiers & os::QUAL_REPEAT) )
					GetWindow()->HandleMessage( new os::Message(M_MENU_FIND_FIND) );
				return;
			}
			case 'r':	// Replace (Ctrl+R)
			{
				if( !(nQualifiers & os::QUAL_REPEAT) )
					GetWindow()->HandleMessage( new os::Message(M_MENU_FIND_REPLACE) );
				return;
			}
			case 'g':	// Goto Line (Ctrl+G)
			{
				if( !(nQualifiers & os::QUAL_REPEAT) )
					GetWindow()->HandleMessage( new os::Message(M_MENU_FIND_GOTO) );
				return;
			}
			case 'q':	// Exit (Ctrl+Q)
			{
				if( !(nQualifiers & os::QUAL_REPEAT) )
					GetWindow()->HandleMessage( new os::Message(M_MENU_APP_QUIT) );
				return;
			}
			default: {}
				/* fall through to cv::CodeView::KeyDown() */
		}
	}

	cv::CodeView::KeyDown( pzString, pzRawString, nQualifiers );
}

int EditView::Find( os::String zOriginal, os::String zFind, bool bCaseSensitive, int nStartPosition )
{
// bCaseSensitive == True, do a case sensitive search
// bCaseSensitive == False, do a case insensitive search

	unsigned int nPosition1, nPosition2 = 0;

	const char* cOriginal = zOriginal.c_str();
	const char* cFind = zFind.c_str();
	bool bMatched = true;

	nPosition1 = nStartPosition;

	if( bCaseSensitive )
	{
		for( ; nPosition1 <= zOriginal.size(); nPosition1++ )
		{
			if( cOriginal[nPosition1] == cFind[0] )
			{
				bMatched = true;
				for( nPosition2 = 0; nPosition2 < zFind.size(); nPosition2++ )
				{
					if( cOriginal[nPosition1 + nPosition2] != cFind[nPosition2] )
					{
						bMatched = false;
						break;
					}
				}

				if( bMatched )
					return(nPosition1);
			}
		}
	}
	else	// Do a case insensitive search
	{
		for( ; nPosition1 <= zOriginal.size(); nPosition1++ )
		{
			if( tolower( cOriginal[nPosition1] ) == tolower(cFind[0]) )
			{
				bMatched = true;

				for( nPosition2 = 0; nPosition2 < zFind.size(); nPosition2++ )
				{
					if( tolower( cOriginal[nPosition1 + nPosition2] ) != tolower( cFind[nPosition2] ) )
					{
						bMatched = false;
						break;
					}
				}

				if( bMatched )
					return(nPosition1);
			}
		}
	}

	return( -1 );
}

void EditView::Undo()
{
	cv::CodeView::Undo();
}

void EditView::Redo()
{
	cv::CodeView::Redo();
}

int EditView::GetLineCount()
{
//	return TextView::GetBuffer().size();
	return( cv::CodeView::GetLineCount() );
}

os::String EditView::GetLine( int32 nLineNo )
{
//	return TextView::GetBuffer()[nLineNo];
	return( cv::CodeView::GetLine(nLineNo) );
}
